/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse;

import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.create.taskmapping.AbstractTaskMappingCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.create.taskmapping.TaskMappingCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.DecoderDependencyGraph;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.CrossoverRandManyAllocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.CrossoverSingleAllocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping.MutateAllocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping.MutateRedundancyDecrement;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping.MutateRedundancyIncrement;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.StrictTTDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.comm.MessageDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatetaskgraph.InstantiatedTaskMappingDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatetaskgraph.InstantiatedTaskMappingDecoderAcyclic;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IRequestAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm.IMessageRouter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm.ShortestPathRouter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.sched.proc.EDFSchedulerBase;

import com.google.inject.TypeLiteral;

/**
 * This abstract factory is responsible for creating the EvaluatorWrappers that are required to hand
 * over objectives and constraints to the optimization. Here, a differentiation is made for the
 * objectives and constraints. The implementation of the specific create functions must be made
 * within factories that are specific for a execution model (like fail silent, fault detection
 * etc.).
 * 
 * @author diewald
 */
public abstract class TaskMappingFactory<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>> {

	/** Error message for non-implemented operators etc. related to schedules */
	private String SCHED_NOT_IMPL_MSG =
			"Error: This operation is not implemented for the execution model defined in the design space exploration. Please implement the operation for this execution model, or avoid calling this function if it is not necessary.";

	/** Error message for non-implemented operators etc. related to optimization goals */
	private String GOAL_NOT_IMPL_MSG =
			"Error: This optimization goal is not implemented. Please implement the goal, or remove it from your design space exploration model.";

	/** The instance (singleton) of the EvaluatorWrapper factory. */
	protected static TaskMappingFactory<?, ?> instance;

	/** Returns the singleton implementing this abstract factory. */
	public static TaskMappingFactory<?, ?> getInstance() {
		if(instance == null) {
			throw new UnsupportedOperationException(
					"TaskMapping factory in not yet initialized. A factory of a concrete TaskMapping must be created first.");
		}
		return instance;
	}

	/* Construct Creators for the task mapping. */

	/** Creates and returns a {@link TypeLiteral} of the specific encoding of the task mapping. */
	public abstract TypeLiteral<T> createGenotypeTypeLiteral();

	/**
	 * Creates an abstract TaskMappingCreator which maps components that needs to be instatiated
	 * first before they can be deployed onto a platform.
	 * 
	 * @param dse
	 *            design space exploration model defined in the GUI
	 * @param execDepGraph
	 *            Dependency graph defining the execution order.
	 * @return instance of an {@link AbstractTaskMappingCreator}.
	 */
	public AbstractTaskMappingCreator
			createAbstractTaskMappingCreator(DecoderDependencyGraph execDepGraph) {
		return new AbstractTaskMappingCreator();
	}

	/**
	 * Creates a TaskMappingCreator that matches the given execution model.
	 * 
	 * @param execDepGraph
	 *            Dependency graph defining the execution order.
	 * @return instance of a {@link TaskMappingCreator}.
	 */
	// TODO: document execDepGraph param.
	public TaskMappingCreator<S, T> createTaskMappingCreator(DecoderDependencyGraph execDepGraph) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates a TaskMappingCreator that uses an {@link AbstractTaskMappingEncoding} to generate a
	 * Tank Mapping that consists only of tasks instances.
	 * 
	 * @param abstractEnc
	 *            encoding that may contain abstract task which need to be instantiated.
	 * @return instance of a {@link TaskMappingCreator}.
	 */
	// TODO: update doc
	public TaskMappingCreator<S, T>
			createTaskMappingCreator(AbstractTaskMappingEncoding abstractEnc) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/* Construct execution model specific operators. */

	/**
	 * Creates an execution model specific {@link MutateAllocation} operator.
	 * 
	 * @param tasks
	 *            {@link ITaskAdapter} that shall be allocated to {@link IRequestAdapter}s.
	 * @return Mutation Operator for Allocations.
	 */
	public MutateAllocation<? extends TaskMappingEntry<?, ?>, ? extends TaskMappingEncoding<?, ?>>
			createMutateAllocationOperator(Collection<ITaskAdapter<?>> tasks) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates an execution model specific {@link MutateRedundancyDecrement} operator.
	 * 
	 * @param tasks
	 *            {@link ITaskAdapter} that shall be allocated to {@link IRequestAdapter}s.
	 * @return Mutation operator that decreases the redundant instances of tasks.
	 */
	public MutateRedundancyDecrement<?, ?>
			createMutateRedundancyDecrementOperator(Collection<ITaskAdapter<?>> tasks) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates an execution model specific {@link MutateRedundancyIncrement} operator.
	 * 
	 * @param tasks
	 *            {@link ITaskAdapter} that shall be allocated to {@link IRequestAdapter}s.
	 * @return Mutation operator that increases the redundant instances of tasks.
	 */
	public MutateRedundancyIncrement<?, ?>
			createMutateRedundancyIncrementOperator(Collection<ITaskAdapter<?>> tasks) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates an execution model specific {@link CrossoverRandManyAllocation} operator
	 * 
	 * @param rate
	 *            The rate or probability of such that the the operator is applied.
	 * @return Crossover operator that swaps randomly many allocations from one resource with
	 *         another.
	 */
	public CrossoverRandManyAllocation<?, ?>
			createCrossoverRandManyAllocationOperator(double rate) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates an execution model specific {@link CrossoverRandManyAllocation} operator
	 * 
	 * @param rate
	 *            The rate or probability of such that the the operator is applied.
	 * @return Crossover operator that swaps a single allocation from one resource with another.
	 */
	public CrossoverSingleAllocation<?, ?> createCrossoverSingleAllocationOperator(double rate) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates and returns a {@link InstantiatedTaskMappingDecoder}.
	 */
	// TODO: remove these suppress warnings by propagating the generic C.
	@SuppressWarnings({"rawtypes", "unchecked"})
	public InstantiatedTaskMappingDecoder<S> createInstantiatedTaskMappingDecoder() {
		return new InstantiatedTaskMappingDecoder();
	}

	/**
	 * Creates and returns a {@link InstantiatedTaskMappingDecoderAcyclic}.
	 * 
	 * @param systemModelAdapter
	 *            The system model deduced from the logical and technical architecture.
	 * @return An instance on an {@link InstantiatedTaskMappingDecoderAcyclic}.
	 */
	public InstantiatedTaskMappingDecoderAcyclic<S> createAbstractToAcyclicTaskGraphDecoder() {
		throw new UnsupportedOperationException(
				"The Decoder to instantiate an \"abstract\" and acyclic Task Graph has not been defined for current Execution Model.");
	}

	/**
	 * Creates and returns a {@link MessageDecoder}.
	 * 
	 * @param systemModelAdapter
	 *            The system model deduced from the logical and technical architecture.
	 * @return An instance of a {@link MessageDecoder}.
	 */
	public MessageDecoder<S, T> createMessageDecoder() {
		throw new UnsupportedOperationException(
				"The Decoder which generates messages from task mappings has not been defined for current Execution Model.");
	}

	/**
	 * Creates a scheduling strategy-"independent" decoder. Here, the basic time architecture is to
	 * be defined, e.g. time triggered scheduling.
	 * 
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture.
	 * @param explorationDefinition
	 *            The model of the DSE; This information can be useful for the decoders to optimize
	 *            the decoding strategy.
	 * @return instance of a {@link StrictTTDecoder}
	 */
	public StrictTTDecoder<S, T> createStrictTTDecoder() {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates a scheduling strategy-"independent" decoder for task mapping encodings. It provides a
	 * {@link ShortestPathRouter} to find the shortest communication path between two communicating
	 * instantiations of an {@link ITaskAdapter} for a given platform.
	 */
	public IMessageRouter createShortestPathRouter() {
		return new ShortestPathRouter();
	}

	/**
	 * Creates an basic EDF scheduler that is used to decode task mappings into schedules for the
	 * given execution model
	 * 
	 * @param expSpec
	 *            DSE goal specification used within the exploration, useful for scheduling
	 * @return instance of the EDF schedule decoder.
	 */
	public EDFSchedulerBase<?, ?> createEDFScheduler() {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}
}
