/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.problem.CompositeProblemModuleBase;
import org.fortiss.af3.exploration.alg.graph.IDependencyModule;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Decoder;

/**
 * {@link Decoder}s for {@link ComposableGenotype}s must implement this interface and be registered
 * as part of a sub problem to the {@link CompositeProblemModuleBase}, such that the sub problem is
 * solved automatically.
 * 
 * @author diewald
 */
public interface ComposableDecoder<G extends ComposableGenotype, P extends Phenotype>
		extends IDependencyModule<IExplorationEncoding, P> {

	/**
	 * Returns the most relevant {@link Genotype} used for the Decoder whereas other dependencies
	 * declare the <i>additionally</i> required information.
	 */
	Class<?> getPrimaryInputType();

	/**
	 * Passes the {@link ComposableGenotype} to the concrete decode method of the concrete
	 * {@link ComposableDecoder} and returns the resulting {@link Phenotype}.
	 * <p>
	 * <b>Implementation note:</b> <br>
	 * The body of the method typically looks as follows: <br>
	 * {@code return decode((ConecreteGenotype)genotype);}
	 * 
	 * @genotype The concrete {@link ComposableGenotype} that shall be decoded.
	 * 
	 * @returns The concrete {@link Phenotype} resulting form the decoding.
	 * 
	 * @throws DecodingException
	 *             If decoding the given {@link Genotype} is not possible.
	 */
	P decodeComposable(G genotype, FlatPhenotypeMap<Phenotype> decodedPhenotypes)
			throws DecodingException;

	/**
	 * Updates the given {@code outGenotype} using the information encoded in the {@code inGenotype}
	 * . It is used for <i>hybrid Phenotypes</i> that are also {@link Genotype} and for which
	 * existing information shall not be destroyed by creating a completely new {@link Phenotype}/
	 * {@link Genotype}.
	 * 
	 * @param inGenotype
	 *            Input {@link Genotype} that is used to update the {@code outGenotype}.
	 * @param outGenotype
	 *            {@link Genotype} to be updated.
	 * @param decodedPhenotypes
	 *            Map of already decoded {@link Phenotype}s.
	 * @return The updated {@link Phenotype}.
	 * @throws DecodingException
	 *             If updating the given {@link Genotype} is not possible.
	 */
	P updateComposable(G inGenotype, P outGenotype, FlatPhenotypeMap<Phenotype> decodedPhenotypes)
			throws DecodingException;
}
