/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.evaluator.ArchExpSubEvaluator;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.dsl.model.MOEAExplorationExpression;
import org.fortiss.af3.exploration.dsl.model.expressions.Expression;
import org.fortiss.af3.exploration.dsl.model.operators.OptimizationDirection;
import org.fortiss.af3.exploration.dsl.model.operators.OptimizationDirectionEnum;
import org.fortiss.af3.exploration.dsl_v2.model.expression.IExpression;
import org.fortiss.af3.exploration.dsl_v2.model.function.Maximize;
import org.fortiss.af3.exploration.dsl_v2.model.function.Minimize;
import org.fortiss.af3.exploration.model.ExplorationObjective;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Objective;
import org.opt4j.core.Objective.Sign;
import org.opt4j.core.Objectives;
import org.opt4j.core.Value;
import org.opt4j.core.problem.Evaluator;

import com.google.common.collect.ClassToInstanceMap;

/**
 * Opt4J specific goal evaluator wrapper that delegates the actual evaluation to a strategy object
 * of type {@link ExplorationTargetEvaluatorBase}
 * 
 * @author huang
 */
public class EvaluatorWrapper<P, RT, T extends ExplorationTarget<RT>>
		implements ArchExpSubEvaluator<P>, Evaluator<P> {

	/** Objective to track evaluation of this {@link ExplorationTarget} */
	private Objective objective;

	/** Actual underlying evaluator */
	private IExplorationTargetEvaluator<P, RT, T> evaluator;

	/** Goal specification */
	private T explorationTarget;

	/** Contains the priority of the {@link ExplorationTarget} of the {@link EvaluatorWrapper}. */
	private int targetPriority;

	/** Phenotypes, i.e. (sub-)solutions, to which this operator may be applied to. */
	private Class<P> targetPhenotypeType;

	/** Constructor that creates an evaluator wrapper for a given {@link ExplorationTarget}. */
	public EvaluatorWrapper(T goal, IExplorationTargetEvaluator<P, RT, T> evaluator, int priority,
			Class<? extends P> primaryInputType) {

		// Construct EA objectives and register it in objective map. Do not give an
		// "optimization direction" (min/max) in case of constraints. They default to minimization.
		OptimizationDirectionEnum direction = OptimizationDirectionEnum.MIN;
		// TODO: Remove after deprecation of the MoeaExpression. Check first to handle this special
		// case.
		if(goal instanceof MOEAExplorationExpression) {
			Expression moExpr = ((MOEAExplorationExpression)goal).getMoeaExpression();
			if(moExpr instanceof OptimizationDirection) {
				direction = ((OptimizationDirection)moExpr).getOperator();
			}
		} else if(goal instanceof ExplorationObjective) {
			// TODO: Reconsider the wrapping / passing to the ExpressionEvaluators. Probably, the
			// evaluators only need to be aware of the systemModelAdapter and the exploration
			// expression.
			IExpression expr = goal.getExpression();
			if(expr instanceof Minimize) {
				direction = OptimizationDirectionEnum.MIN;
			} else if(expr instanceof Maximize) {
				direction = OptimizationDirectionEnum.MAX;
			}
		}

		String objectiveName =
				goal.getName() == null ? goal.getClass().toString() + Integer.toString(goal.getId())
						: goal.getName() + Integer.toString(goal.getId());
		objective = new Objective(objectiveName,
				direction == OptimizationDirectionEnum.MIN ? Sign.MIN : Sign.MAX);

		this.explorationTarget = goal;
		this.evaluator = evaluator;
		this.targetPriority = priority;
		@SuppressWarnings("unchecked") Class<P> targetPhenotypeType = (Class<P>)primaryInputType;
		this.targetPhenotypeType = targetPhenotypeType;
	}

	/**
	 * Returns the priority of this evaluator (based on the priority specified for the underlying
	 * goal)
	 */
	public int getPriority() {
		return targetPriority;
	}

	/** {@inheritDoc} */
	@Override
	public <CP extends Phenotype> Objectives evaluateComposable(Class<CP> clazz, CP phenotype,
			FlatPhenotypeMap<Phenotype> phenotypes, ClassToInstanceMap<EObject> transformedModels) {
		Objectives objectives = new Objectives();

		// TODO: Do we really need the evalu
		@SuppressWarnings("unchecked") P castedPhenotype = (P)phenotype;

		Value<?> evalResult = new DoubleValue(Double.MAX_VALUE);
		try {
			evalResult = evaluator.evaluateGoal(castedPhenotype, phenotypes, transformedModels);
		} catch(Exception e) {
			// The Opt4J framework only allows us to throw runtime exceptions.
			throw new RuntimeException(new ExplorationException("The evaluatior " +
					this.toString() + " failed to evaluate the ExplorationTarget " +
					explorationTarget.getName() + ".", e));
		}

		objectives.add(objective, evalResult);

		return objectives;
	}

	/** {@inheritDoc} */
	@Override
	public Objectives evaluate(P phenotype) {
		// TODO: Investigate when this method is called. Should be never the case...
		assert (false);
		Objectives objectives = new Objectives();

		Value<?> evalResult = new DoubleValue(Double.MAX_VALUE);
		try {
			evalResult = evaluator.evaluateGoal(phenotype, null, null);
		} catch(Exception e) {
			// The Opt4J framework only allows us to throw runtime exceptions.
			throw new RuntimeException(new ExplorationException("The evaluatior " +
					this.toString() + " failed to evaluate the ExplorationTarget " +
					explorationTarget.getName() + ".", e));
		}

		objectives.add(objective, evalResult);

		return objectives;
	}

	/** Returns the type of the goal this evaluator is built for. */
	@SuppressWarnings("unchecked")
	public Class<T> getTargetType() {
		return (Class<T>)explorationTarget.getClass();
	}

	/** Returns the underlying (actual) evaluator. */
	public IExplorationTargetEvaluator<P, RT, T> getEvaluator() {
		return evaluator;
	}

	/** Returns the objective associated with the wrapped {@link Evaluator}. */
	public Objective getObjective() {
		return objective;
	}

	/** Returns the class of the phenotype to */
	public Class<P> getTargetPhenotypeType() {
		return targetPhenotypeType;
	}
}
