/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction;

import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.comm.constraint.InternalIsolatedCommunicationSet;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.UpdatesEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.Decodes;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.Genotyped;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.decode.ConstraintGenerationDecoder;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.opt4j.core.Genotype;

import com.google.inject.Provides;
import com.google.inject.name.Named;

/**
 * Decodes a {@link SafetyFunctionArchEncoding} into an {@link AbstractTaskMappingEncoding}.
 * 
 * @author diewald
 */
public class SFMappingDecoder<C> extends DecoderModule<AbstractTaskMappingEncoding>
		implements ConstraintGenerationDecoder {

	/** {@inheritDoc} */
	@Override
	public AbstractTaskMappingEncoding decode(Genotype genotype) {
		return null;
	}

	/**
	 * Decodes a {@link SafetyFunctionArchEncoding} into an {@link AbstractTaskMappingEncoding}
	 * where allocations are added or removed such that all Tasks of the (replicated) safety
	 * channels and their voters are mapped to the target platform. To calculate these updates, the
	 * {@link AbstractTaskMappingEncoding} from a previous iteration is required.
	 * 
	 * @param sfaEnc
	 *            Encoding of the architectures of safety functions present in the system.
	 * @param atmEnc
	 *            Mapping of tasks to the target execution platform.
	 * @return Updated / Aligned task mapping.
	 * @throws DecodingException
	 *             if the creation of some mapping entry fails.
	 */
	@Provides
	@Named("SFMappingDecoder")
	@Decodes
	@UpdatesEncoding
	public AbstractTaskMappingEncoding decode(@Genotyped SafetyFunctionArchEncoding sfaEnc,
			@Genotyped AbstractTaskMappingEncoding atmEnc) throws DecodingException {

		// TODO: replace with a dynamic mechanism for the retrieval of deployable components.
		for(SafetyFunctionArchEntry sFAEntry : sfaEnc.getAllEntries()) {
			boolean diagPresent = sFAEntry.isDiagnosicUnitPresent();
			int channelNumber = sFAEntry.getNumChannels();

			// TODO: Encapsulate in the encoding / adapter --> Raw cast.
			@SuppressWarnings("unchecked") Collection<ITaskAdapter<?>> sfComps =
					(Collection<ITaskAdapter<?>>)(Collection<?>)sFAEntry.getSafetyFunctionAdapter()
							.getRealizingComponents();
			@SuppressWarnings("unchecked") Collection<ITaskAdapter<?>> diagComps =
					(Collection<ITaskAdapter<?>>)(Collection<?>)sFAEntry.getSafetyFunctionAdapter()
							.getDiagnosticComponents();
			ITaskAdapter<?> outComponent = sFAEntry.getSafetyFunctionAdapter().getOutputComponent();

			// We assume that the entry number must be equal for all realizing components of a
			// safety function.
			Collection<AbstractTaskMappingEntry> sfEntries =
					atmEnc.getMappingEntriesOf(sfComps.stream().findAny().get());
			int entryNumber = sfEntries.size();

			// TODO: use a more generic solution to instantiate these tyoes of constraints during
			// the creation phase.
			Set<IMappingEntry<?, ?>> existingEntries = new HashSet<>();
			if(atmEnc.getConstraintsOf(InternalIsolatedCommunicationSet.class).isEmpty()) {
				for(ITaskAdapter<?> comp : sFAEntry.getSafetyFunctionAdapter().getAllComponents()) {
					Collection<AbstractTaskMappingEntry> curEntry =
							atmEnc.getMappingEntriesOf(comp);
					assert (curEntry.size() <= 1);
					existingEntries.addAll(curEntry);
				}
				atmEnc.putConstraint(InternalIsolatedCommunicationSet.class,
						new InternalIsolatedCommunicationSet(existingEntries));
			}

			// In case that diagnostic units are not present in the *previous* encoding, we need to
			// add some more.
			Collection<AbstractTaskMappingEntry> diagEntries =
					atmEnc.getMappingEntriesOf(diagComps.stream().findAny().get());
			if(diagPresent && diagEntries != null && diagEntries.isEmpty()) {
				try {
					addMissingDiagUnits(atmEnc, diagComps, entryNumber);
				} catch(ExplorationException e) {
					throw new DecodingException(sfaEnc, this,
							"Error adding diagnostic components to match the number of safety channels.",
							e);
				}
			}

			// Match the encoding with the selected channel number.
			for(int i = entryNumber; i < channelNumber; i++) {
				try {
					Set<IMappingEntry<?, ?>> createdEntries = new HashSet<>();
					createdEntries.addAll(addSFComponents(atmEnc, sfComps));
					if(diagPresent) {
						createdEntries.addAll(addSFComponents(atmEnc, diagComps));
					}
					// TODO: for each channel, the output component is "replicated" here.
					createdEntries.addAll(addSFComponents(atmEnc, Arrays.asList(outComponent)));

					atmEnc.putConstraint(InternalIsolatedCommunicationSet.class,
							new InternalIsolatedCommunicationSet(createdEntries));
				} catch(ExplorationException e) {
					throw new DecodingException(sfaEnc, this,
							"Error adding safety components to match the number of safety channels.",
							null);
				}
			}
			List<InternalIsolatedCommunicationSet> isoCommSet = new ArrayList<>(
					atmEnc.getConstraintsOf(InternalIsolatedCommunicationSet.class));
			for(int i = entryNumber - 1; i > channelNumber; i--) {
				Collection<IMappingEntry<?, ?>> rmEntries =
						isoCommSet.get(i).getIsolatedCommunicationSet();

				removeSFEntries(atmEnc, rmEntries);
				removeMatchingConstraint(atmEnc, getFirst(rmEntries, (s) -> true).orElse(null));
			}

			if(!diagPresent) {
				// If no diagnostic unit is present in the encoding, remove its entire task.
				removeDiagUnits(atmEnc, diagComps);
			}
		}

		return atmEnc;
	}

	/**
	 * Adds entries for diagnostic units in the task mapping up to the given number. To be used when
	 * the encoding previously did not contain diagnostic units.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one {@link ITaskAdapter}.
	 */
	private void addMissingDiagUnits(AbstractTaskMappingEncoding outGenotype,
			Collection<ITaskAdapter<?>> diagComps, int entryNumber) throws ExplorationException {
		List<InternalIsolatedCommunicationSet> isoComm = new ArrayList<>(
				outGenotype.getConstraintsOf(InternalIsolatedCommunicationSet.class));
		for(int i = 0; i < entryNumber; i++) {
			Set<IMappingEntry<?, ?>> createdEntries = addSFComponents(outGenotype, diagComps);
			InternalIsolatedCommunicationSet curConstr = isoComm.get(i);
			curConstr.addToIsolatedCommunicationSet(createdEntries);
		}
	}

	/** Entirely removes the diagnostic units of a safety function from a given mapping. */
	private void removeDiagUnits(AbstractTaskMappingEncoding outGenotype,
			Collection<ITaskAdapter<?>> diagComps) {
		for(ITaskAdapter<?> diagComp : diagComps) {
			Collection<AbstractTaskMappingEntry> removedEntries = outGenotype.removeTask(diagComp);
			Collection<InternalIsolatedCommunicationSet> isoCommConstrs =
					outGenotype.getConstraintsOf(InternalIsolatedCommunicationSet.class);
			if(isoCommConstrs != null) {
				// During the first iteration, Diagnostic units may be removed but no constraints
				// were generated for them yet, hence check for null here.
				isoCommConstrs.forEach(c -> c.removeFromIsolatedCommunicationSet(removedEntries));
			}
		}
	}

	/**
	 * Adds one random entry to the task mapping for each of the given {@link ITaskAdapter}s.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one {@link ITaskAdapter}.
	 */
	private Set<IMappingEntry<?, ?>> addSFComponents(AbstractTaskMappingEncoding outGenotype,
			Collection<ITaskAdapter<?>> sfComps) throws ExplorationException {
		Set<IMappingEntry<?, ?>> createdEntries = new HashSet<>();
		if(sfComps != null) {
			for(ITaskAdapter<?> curSFComp : sfComps) {
				createdEntries.add(outGenotype.createAndAddRandomEntry(curSFComp));
			}
		}

		return createdEntries;
	}

	/** Removes the given {@link IMappingEntry}s from the {@link AbstractTaskMappingEncoding}. */
	private void removeSFEntries(AbstractTaskMappingEncoding outGenotype,
			Collection<IMappingEntry<?, ?>> sfEntries) {
		for(IMappingEntry<?, ?> sfEntry : sfEntries) {
			outGenotype.removeEntry((AbstractTaskMappingEntry)sfEntry);
		}
	}

	/**
	 * Returns the first found {@link InternalIsolatedCommunicationSet} constraint that includes
	 * the given {@link IMappingEntry}. If none is found {@code null} is returned.
	 */
	private InternalIsolatedCommunicationSet getMatchingIsolationCommConstraint(
			AbstractTaskMappingEncoding outGenotype, IMappingEntry<?, ?> entry) {
		InternalIsolatedCommunicationSet matchingConstraint = null;
		Collection<InternalIsolatedCommunicationSet> isoComm =
				outGenotype.getConstraintsOf(InternalIsolatedCommunicationSet.class);
		for(InternalIsolatedCommunicationSet constr : isoComm) {
			if(constr.getIsolatedCommunicationSet().contains(entry)) {
				matchingConstraint = constr;
				break;
			}
		}

		return matchingConstraint;
	}

	/**
	 * Removes the first found {@link InternalIsolatedCommunicationSet} constraint that includes the
	 * given {@link IMappingEntry} from the given {@link AbstractTaskMappingEncoding}.
	 */
	private void removeMatchingConstraint(AbstractTaskMappingEncoding outGenotype,
			IMappingEntry<?, ?> entry) {
		InternalIsolatedCommunicationSet rmConstr =
				getMatchingIsolationCommConstraint(outGenotype, entry);
		assert rmConstr != null;
		outGenotype.getConstraintsOf(InternalIsolatedCommunicationSet.class).remove(rmConstr);
	}
}
