/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel;

import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.TaskMappingFactory;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.create.taskmapping.FailSilentTaskMappingCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.create.taskmapping.TaskMappingCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.DecoderDependencyGraph;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.CrossoverRandManyAllocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.CrossoverSingleAllocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping.MutateAllocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping.MutateRedundancyDecrement;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping.MutateRedundancyIncrement;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.StrictTTDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.comm.MessageDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatetaskgraph.InstantiatedTaskMappingDecoderAcyclic;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.sched.proc.EDFSchedulerBase;
import org.fortiss.af3.exploration.alg.dse.sysmodel.sched.proc.FailSilentEDFScheduler;

import com.google.inject.TypeLiteral;

/**
 * This class implements the creation of the EavaluatorWrapper for the fail silent execution model.
 * 
 * @author diewald
 */
public class FailSilentExecModelFactory
		extends TaskMappingFactory<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding> {

	/** The instance (singleton) of the EvaluatorWrapper factory. */
	public static TaskMappingFactory<?, ?> getInstance() {
		if(instance == null) {
			instance = new FailSilentExecModelFactory();
		}
		return instance;
	}

	/** {@inheritDoc} */
	@Override
	public TypeLiteral<FailSilentTaskMappingEncoding> createGenotypeTypeLiteral() {
		return new TypeLiteral<FailSilentTaskMappingEncoding>() {
			// Anonymous.
		};
	}

	/** Creates a TaskMappingCreator that matches the given execution model. */
	@Override
	public TaskMappingCreator<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding>
			createTaskMappingCreator(DecoderDependencyGraph execDepGraph) {
		return new FailSilentTaskMappingCreator();
	}

	/* Construct execution model specific operators. */

	/** {@inheritDoc} */
	@Override
	public MutateAllocation<?, ?>
			createMutateAllocationOperator(Collection<ITaskAdapter<?>> tasks) {
		return new MutateAllocation<AbstractTaskMappingEntry, AbstractTaskMappingEncoding>(tasks);
	}

	/** {@inheritDoc} */
	@Override
	public MutateRedundancyDecrement<?, ?>
			createMutateRedundancyDecrementOperator(Collection<ITaskAdapter<?>> tasks) {
		return new MutateRedundancyDecrement<AbstractTaskMappingEntry, AbstractTaskMappingEncoding>(
				tasks);
	}

	/** {@inheritDoc} */
	@Override
	public MutateRedundancyIncrement<?, ?>
			createMutateRedundancyIncrementOperator(Collection<ITaskAdapter<?>> tasks) {
		return new MutateRedundancyIncrement<AbstractTaskMappingEntry, AbstractTaskMappingEncoding>(
				tasks);
	}

	/** {@inheritDoc} */
	@Override
	public CrossoverRandManyAllocation<?, ?>
			createCrossoverRandManyAllocationOperator(double rate) {
		return new CrossoverRandManyAllocation<AbstractTaskMappingEntry, AbstractTaskMappingEncoding>(
				rate);
	}

	/** {@inheritDoc} */
	@Override
	public CrossoverSingleAllocation<?, ?> createCrossoverSingleAllocationOperator(double rate) {
		return new CrossoverSingleAllocation<AbstractTaskMappingEntry, AbstractTaskMappingEncoding>(
				rate);
	}

	/* Construct Decoders for the genotypes, aka task mappings. */

	/** {@inheritDoc} */
	@SuppressWarnings({"rawtypes", "unchecked"})
	@Override
	// TODO: identify why a warning is generated here.
	public InstantiatedTaskMappingDecoderAcyclic<FailSilentTaskMappingEntry>
			createAbstractToAcyclicTaskGraphDecoder() {
		return new InstantiatedTaskMappingDecoderAcyclic();
	}

	/** {@inheritDoc} */
	@Override
	public MessageDecoder<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding>
			createMessageDecoder() {
		return new MessageDecoder<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding>();
	}

	/** {@inheritDoc} */
	@Override
	public StrictTTDecoder<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding>
			createStrictTTDecoder() {
		return new StrictTTDecoder<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding>();
	}

	/**
	 * Creates an basic EDF scheduler that is used to decode task mappings into schedules for the
	 * given execution model
	 */
	@Override
	public EDFSchedulerBase<?, ?> createEDFScheduler() {
		return new FailSilentEDFScheduler();
	}
}
