/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service;

import java.util.Collection;
import java.util.Comparator;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.ComposableEvaluator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.dse.module.IEvaluatorModule;
import org.fortiss.af3.exploration.model.ExplorationTarget;

/**
 * {@link IExplorationService} interface to register evaluators for the DSE. See
 * {@link IExplorationTargetEvaluator} for the interface that needs to be implemented by concrete
 * evaluator modules.
 * 
 * @author diewald
 */
public interface IExplorationEvaluatationService extends IExplorationService {

	/**
	 * Registers the given {@link EvaluatorModule} with the {@link IExplorationEvaluatationService}.
	 * It is added to the set of exploration modules of the {@link IExplorationFeatureService}.
	 * 
	 * @param evaluatorModule
	 *            {@link EvaluatorModule} to be registered.
	 */
	void registerEvaluator(EvaluatorModule<?, ?, ?> evaluatorModule);

	/**
	 * Returns an instance of the {@link EvaluatorModule} that is used to evaluate a solution (
	 * {@link Phenotype}) w.r.t. the given target.
	 * 
	 * @param explorationTarget
	 *            {@link ExplorationTarget} to be evaluated.
	 * @return the corresponding {@link EvaluatorModule}.
	 */
	<R, T extends ExplorationTarget<R>> IEvaluatorModule<R, T, ?>
			getEvaluatorInstance(T explorationTarget);

	/**
	 * Returns instances of the {@link EvaluatorModule}s that are used to evaluate a solution (
	 * {@link Phenotype}) w.r.t. the given targets.
	 * 
	 * @param explorationTarget
	 *            List of {@link ExplorationTarget} to be evaluated.
	 * @param prioComparator
	 *            {@link Comparator} that defines the order of execution of the
	 *            {@link IEvaluatorModule}s.
	 * @return List of the corresponding {@link EvaluatorModule}.
	 */
	<R, T extends ExplorationTarget<R>> Set<IEvaluatorModule<?, ?, ?>> getEvaluatorInstances(
			Set<? extends ExplorationTarget<?>> explorationTarget,
			Comparator<IEvaluatorModule<?, ?, ?>> prioComparator);

	/**
	 * Returns the collection of input model types that are required to execute the given set of
	 * {@link ComposableEvaluator}s. The models corresponding to these types can be retrieved via
	 * the {@link IExplorationTransformationService}.
	 * 
	 * @param evaluators
	 *            Set of evaluators that shall be executed.
	 * @return collection of model element types which must be provided through the transformation
	 *         service (or directly) such that evaluator gets all the required inputs.
	 */
	Collection<Class<?>>
			getRequiredTransformedModels(Set<IExplorationTargetEvaluator<?, ?, ?>> evaluators);
}
